@@ -4,25 +4,7 @@
 #
 # Author:      Andras Wacha
 #
-# Created:     2010/02/22
-# RCS-ID:      $Id: B1io.py $
-# Copyright:   (c) 2010
-# Licence:     GPLv2
 #-----------------------------------------------------------------------------
-#       This program is free software; you can redistribute it and/or modify
-#       it under the terms of the GNU General Public License as published by
-#       the Free Software Foundation; either version 2 of the License, or
-#       (at your option) any later version.
-#       
-#       This program is distributed in the hope that it will be useful,
-#       but WITHOUT ANY WARRANTY; without even the implied warranty of
-#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-#       GNU General Public License for more details.
-#       
-#       You should have received a copy of the GNU General Public License
-#       along with this program; if not, write to the Free Software
-#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
-#       MA 02110-1301, USA.
 
 import pylab
 import numpy as np
@@ -36,7 +18,7 @@ import utils
 import re
 import warnings
 import fitting
-
+import datetime
 
 from c_B1io import cbfdecompress
 HC=12398.419 #Planck's constant times speed of light, in eV*Angstrom units
@@ -47,6 +29,63 @@ try:
 except ImportError:
     warnings.warn('Cannot import module Image (Python Imaging Library). Only Pilatus300k and Gabriel images can be loaded (Pilatus100k, 1M etc. NOT).')
 
+def readyellowsubmarine(nameformat,fsns=None,dirs='.'):
+    """
+    """
+    if fsns is None:
+        filenames=[nameformat]
+    else:
+        filenames=[nameformat%f for f in fsns]
+    if type(dirs)!=type([]) and type(dirs)!=type(tuple()):
+        dirs=[dirs]
+    datas=[]
+    params=[]
+    for fn in filenames:
+        for d in dirs:
+            try:
+                f=open(os.path.join(d,fn),'r')
+            except IOError:
+                continue
+            try:
+                s=f.read()            
+                f.close()
+                par={}
+                par['FSN']=int(s[2:6])
+                par['Owner']=s[6:0x18].split()[0]
+                par['Title']='_'.join(s[6:0x18].split()[1:])
+                par['MeasTime']=long(s[0x18:0x1e])
+                par['Monitor']=long(s[0x1e:0x26])
+                par['Day']=int(s[0x26:0x28])
+                par['Month']=int(s[0x29:0x2b])
+                par['Year']=int(s[0x2c:0x30])
+                par['Hour']=int(s[0x30:0x32])
+                par['Minute']=int(s[0x33:0x35])
+                par['Second']=int(s[0x36:0x38])
+                par['PosSample']=int(s[0x60:0x65])
+                par['PosBS']=int(s[0x5b:0x60])
+                par['PosDetector']=int(s[0x56:0x5b])
+                par['max']=long(s[0x38:0x3d])
+                par['selector_speed']=long(s[0x3d:0x42])
+                par['wavelength']=long(s[0x42:0x44])
+                par['Dist_Ech_det']=long(s[0x44:0x49])
+                par['comments']=s[0x6d:0x100]
+                par['sum']=long(s[0x65:0x6d])
+                par['BeamPosX']=float(s[0x49:0x4d])
+                par['BeamPosY']=float(s[0x4d:0x51])
+                par['AngleBase']=float(s[0x51:0x56])
+                par['Datetime']=datetime.datetime(par['Year'],par['Month'],par['Day'],par['Hour'],par['Minute'],par['Second'])
+                
+                params.append(par)
+                datas.append(np.fromstring(s[0x100:],np.uint16).astype(np.double).reshape((64,64)))
+                break
+            except ValueError:
+                print "File %s is invalid! Skipping."%fn
+                continue
+    if fsns is None:
+        return datas[0],params[0]
+    else:
+        return datas,params
+
 
 def readcbf(name):
     """Read a cbf (crystallographic binary format) file from a Dectris Pilatus detector.
@@ -278,17 +317,19 @@ def bdf_read(filename):
             bdf['CT'][t_list[j]]=t_value[j]
     fid.close()
     return bdf
+
 def readasa(basename,dirs=[]):
     """Load SAXS/WAXS measurement files from ASA *.INF, *.P00 and *.E00 files.
     
     Input:
-        basename: the basename (without extension) of the files
+        basename: the basename (without extension) of the files. Can also be a
+            list of strings
         dirs: list of directories (or just a single directory) to search files
             in. P00, INF and E00 should reside in the same directory.
     Output:
-        An ASA dictionary of the following fields:
-            position: the counts for each pixel (numpy array)
-            energy: the energy spectrum (numpy array)
+        An ASA dictionary (or a list of them) with the following fields:
+            position: the counts for each pixel (numpy array), in cps
+            energy: the energy spectrum (numpy array), in cps
             params: parameter dictionary. It has the following fields:
                 Month: The month of the measurement
                 Day: The day of the measurement
@@ -307,93 +348,110 @@ def readasa(basename,dirs=[]):
                 Realtime: real time in seconds
                 Livetime: live time in seconds
             pixels: the pixel numbers.
+            poserror: estimated error of the position (cps)
+            energyerror: estimated error of the energy (cps)
     """
     if type(dirs)==type(''):
         dirs=[dirs]
     if len(dirs)==0:
         dirs=['.']
-    for d in dirs:
-        try:
-            p00=np.loadtxt(os.path.join(d,'%s.P00' % basename))
-        except IOError:
+    if type(basename)==type(''):
+        basenames=[basename]
+        basename_scalar=True
+    else:
+        basenames=basename
+        basename_scalar=False
+    ret=[]
+    for basename in basenames:
+        for d in dirs:
             try:
-                p00=np.loadtxt(os.path.join(d,'%s.p00' % basename))
-            except:
-                p00=None
-        if p00 is not None:
-            p00=p00[1:] # cut the leading -1
-        try:
-            e00=np.loadtxt(os.path.join(d,'%s.E00' % basename))
-        except IOError:
-            try:
-                e00=pylab.loadtxt(os.path.join(d,'%s.e00' % basename))
-            except:
-                e00=None
-        if e00 is not None:
-            e00=e00[1:] # cut the leading -1
-        try:
-            inffile=open(os.path.join(d,'%s.inf' % basename))
-        except IOError:
-            try:
-                inffile=open(os.path.join(d,'%s.Inf' % basename))
+                p00=np.loadtxt(os.path.join(d,'%s.P00' % basename))
             except IOError:
                 try:
-                    inffile=open(os.path.join(d,'%s.INF' % basename))
+                    p00=np.loadtxt(os.path.join(d,'%s.p00' % basename))
                 except:
-                    inffile=None
-                    params=None
-        if (p00 is not None) and (e00 is not None) and (inffile is not None):
-            break
-        else:
-            p00=None
-            e00=None
-            inffile=None
-    if (p00 is None) or (e00 is None) or (inffile is None):
-        print "Cannot find every file (*.P00, *.INF, *.E00) for sample %s in any directory" %basename
-        return None
-    if inffile is not None:
-        params={}
-        l1=inffile.readlines()
-        l=[]
-        for line in l1:
-            if len(line.strip())>0:
-                l.append(line)
-        def getdate(str):
+                    p00=None
+            if p00 is not None:
+                p00=p00[1:] # cut the leading -1
             try:
-                month=int(str.split()[0].split('-')[0])
-                day=int(str.split()[0].split('-')[1])
-                year=int(str.split()[0].split('-')[2])
-                hour=int(str.split()[1].split(':')[0])
-                minute=int(str.split()[1].split(':')[1])
-                second=int(str.split()[1].split(':')[2])
-            except:
-                return None
-            return {'Month':month,'Day':day,'Year':year,'Hour':hour,'Minute':minute,'Second':second}
-        if getdate(l[0]) is None:
-            params['Title']=l[0].strip()
-            offset=1
-        else:
-            params['Title']=basename
-            offset=0
-        d=getdate(l[offset])
-        params.update(d)
-        for line in l:
-            if line.strip().startswith('PSD1 Lower Limit'):
-                params['Energywindow_Low']=float(line.strip().split(':')[1].replace(',','.'))
-            elif line.strip().startswith('PSD1 Upper Limit'):
-                params['Energywindow_High']=float(line.strip().split(':')[1].replace(',','.'))
-            elif line.strip().startswith('Realtime'):
-                params['Realtime']=float(line.strip().split(':')[1].split()[0].replace(',','.').replace('\xa0',''))
-            elif line.strip().startswith('Lifetime'):
-                params['Livetime']=float(line.strip().split(':')[1].split()[0].replace(',','.').replace('\xa0',''))
-            elif line.strip().startswith('Lower Limit'):
-                params['Energywindow_Low']=float(line.strip().split(':')[1].replace(',','.'))
-            elif line.strip().startswith('Upper Limit'):
-                params['Energywindow_High']=float(line.strip().split(':')[1].replace(',','.'))
-            elif line.strip().startswith('Stop Condition'):
-                params['Stopcondition']=line.strip().split(':')[1].strip().replace(',','.')
-        params['basename']=basename.split(os.sep)[-1]
-    return {'position':p00,'energy':e00,'params':params,'pixels':pylab.arange(len(p00))}
+                e00=np.loadtxt(os.path.join(d,'%s.E00' % basename))
+            except IOError:
+                try:
+                    e00=pylab.loadtxt(os.path.join(d,'%s.e00' % basename))
+                except:
+                    e00=None
+            if e00 is not None:
+                e00=e00[1:] # cut the leading -1
+            try:
+                inffile=open(os.path.join(d,'%s.inf' % basename))
+            except IOError:
+                try:
+                    inffile=open(os.path.join(d,'%s.Inf' % basename))
+                except IOError:
+                    try:
+                        inffile=open(os.path.join(d,'%s.INF' % basename))
+                    except:
+                        inffile=None
+                        params=None
+            if (p00 is not None) and (e00 is not None) and (inffile is not None):
+                break
+            else:
+                p00=None
+                e00=None
+                inffile=None
+        if (p00 is None) or (e00 is None) or (inffile is None):
+            print "Cannot find every file (*.P00, *.INF, *.E00) for sample %s in any directory" %basename
+            continue
+        if inffile is not None:
+            params={}
+            l1=inffile.readlines()
+            l=[]
+            for line in l1:
+                if len(line.strip())>0:
+                    l.append(line)
+            def getdate(str):
+                try:
+                    month=int(str.split()[0].split('-')[0])
+                    day=int(str.split()[0].split('-')[1])
+                    year=int(str.split()[0].split('-')[2])
+                    hour=int(str.split()[1].split(':')[0])
+                    minute=int(str.split()[1].split(':')[1])
+                    second=int(str.split()[1].split(':')[2])
+                except:
+                    return None
+                return {'Month':month,'Day':day,'Year':year,'Hour':hour,'Minute':minute,'Second':second}
+            if getdate(l[0]) is None:
+                params['Title']=l[0].strip()
+                offset=1
+            else:
+                params['Title']=basename
+                offset=0
+            d=getdate(l[offset])
+            params.update(d)
+            for line in l:
+                if line.strip().startswith('PSD1 Lower Limit'):
+                    params['Energywindow_Low']=float(line.strip().split(':')[1].replace(',','.'))
+                elif line.strip().startswith('PSD1 Upper Limit'):
+                    params['Energywindow_High']=float(line.strip().split(':')[1].replace(',','.'))
+                elif line.strip().startswith('Realtime'):
+                    params['Realtime']=float(line.strip().split(':')[1].split()[0].replace(',','.').replace('\xa0',''))
+                elif line.strip().startswith('Lifetime'):
+                    params['Livetime']=float(line.strip().split(':')[1].split()[0].replace(',','.').replace('\xa0',''))
+                elif line.strip().startswith('Lower Limit'):
+                    params['Energywindow_Low']=float(line.strip().split(':')[1].replace(',','.'))
+                elif line.strip().startswith('Upper Limit'):
+                    params['Energywindow_High']=float(line.strip().split(':')[1].replace(',','.'))
+                elif line.strip().startswith('Stop Condition'):
+                    params['Stopcondition']=line.strip().split(':')[1].strip().replace(',','.')
+            params['basename']=basename.split(os.sep)[-1]
+        ret.append({'position':p00/params['Livetime'],'energy':e00/params['Livetime'],
+                'params':params,'pixels':pylab.arange(len(p00)),
+                'poserror':np.sqrt(p00)/params['Livetime'],
+                'energyerror':np.sqrt(e00)/params['Livetime']})
+    if basename_scalar:
+        return ret[0]
+    else:
+        return ret
 def readheader(filename,fsn=None,fileend=None,dirs=[],quiet=False):
     """Reads header data from measurement files
     
@@ -1134,6 +1192,7 @@ def readlogfile(fsn,dirs=[],norm=True,quiet=False):
         if not filefound and not quiet:
             print 'Cannot find file %s in any of the given directories.' % filebasename
     return params;
+            
 def writelogfile(header,ori,thick,dc,realenergy,distance,mult,errmult,reffsn,
                  thickGC,injectionGC,injectionEB,pixelsize,mode='Pilatus',norm=True):
     """Write logfiles.
@@ -1465,11 +1524,12 @@ def writef1f2(f1f2,filename):
         filename: file name
     """
     pylab.savetxt(filename,f1f2,delimiter='\t')
-def readabt(filename):
+def readabt(filename,dirs='.'):
     """Read abt_*.fio type files.
     
     Input:
         filename: the name of the file.
+        dirs: directories to search for files in
         
     Output:
         A dictionary with the following fields:
@@ -1480,50 +1540,88 @@ def readabt(filename):
             'dataset': a structured array a la numpy, containing the same data
                 as in 'data', but in another representation.
     """
-    try:
-        f=open(filename,'rt');
-    except IOError:
-        print 'Cannot open file %s' % filename
-        return None
-    rows=0;
-    a=f.readline(); rows=rows+1;
-    while a[:2]!='%c' and len(a)>0:
-        a=f.readline();  rows=rows+1;
-    if len(a)<=0:
-        print 'Invalid format: %c not found'
-        f.close()
-        return None
-    a=f.readline(); rows=rows+1;
-    if a[:7]==' ENERGY':
-        mode='Energy'
-    elif a[:4]==' MOT':
-        mode='Motor'
-    else:
-        print 'Unknown scan type: %s' % a
-        f.close()
-        return None
-    f.readline(); rows=rows+1;
-    title=f.readline()[:-1]; rows=rows+1;
-    while a[:2]!='%d' and len(a)>0:
-        a=f.readline(); rows=rows+1;
-    if len(a)<=0:
-        print 'Invalid format: %d not found'
-        f.close()
-        return None
-    columns=[];
-    a=f.readline(); rows=rows+1;
-    while a[:4]==' Col':
-        columns.append(a.split()[2][10:]);
-        a=f.readline(); rows=rows+1;
-        #print a
-    #print a[:4]
-    f.seek(-len(a),1)
-    rows=rows-1;
-    #print rows
-    matrix=np.loadtxt(f)
-    data=np.array([tuple(a) for a in matrix.tolist()], dtype=zip(columns,[np.double]*len(columns)))
-    f.close()
-    return {'title':title,'mode':mode,'columns':columns,'data':matrix,'dataset':data}
+    if type(dirs)==type(''):
+        dirs=[dirs]
+    if len(dirs)==0:
+        dirs=['.']
+    ret={}
+    for d in dirs:
+        try:
+            f=open(os.path.join(d,filename),'rt');
+        except IOError:
+            print 'Cannot open file %s' % filename
+            continue
+        # now the file is open
+        lines=f.readlines()
+        #prune comment lines starting with an exclamation mark (!).
+        i=0
+        while i<len(lines):
+            lines[i]=lines[i].strip()
+            if lines[i].startswith('!') or len(lines[i])==0:
+                lines.pop(i)
+                i-=1
+            i+=1
+        # find the parameter part
+        ret['params']={}
+        idx=lines.index('%p')+1
+        while idx<len(lines) and (not lines[idx].startswith('%')):
+            ls=lines[idx].split('=')
+            if len(ls)==2:
+                ret['params'][ls[0].strip()]=float(ls[1].strip())
+            idx+=1
+        # find the comment part
+        idx=lines.index('%c')
+        # first comment line is like: MOT12-Scan started at 21-Sep-2009 13:43:56, ended 13:47:53
+        l=lines[idx+1]
+        if l.startswith('MOT'):
+            ret['mode']='Motor'
+        elif l.startswith('ENERGY'):
+            ret['mode']='Energy'
+        else:
+            print 'Unknown scan type!'
+            return None
+        # find the string containing the start time in dd-mmm-yyyy hh:mm:ss format
+        str=l[(l.index('started at ')+len('started at ')):l.index(', ended')]
+        date,time1=str.split(' ')
+        ret['params']['day'],ret['params']['month'],ret['params']['year']=date.split('-')
+        ret['params']['day']=int(ret['params']['day'])
+        ret['params']['year']=int(ret['params']['year'])
+        ret['params']['hourstart'],ret['params']['minutestart'],ret['params']['secondstart']=[int(x) for x in time1.split(':')]
+        str=l[l.index('ended')+len('ended '):]
+        ret['params']['hourend'],ret['params']['minuteend'],ret['params']['secondend']=[int(x) for x in str.split(':')]
+        
+        l=lines[idx+2]
+        if l.startswith('Name:'):
+            ret['name']=l.split(':')[1].split()[0]
+        else:
+            raise ValueError('File %s is invalid!' % filename)
+        
+        l=lines[idx+3]
+        if l.startswith('Counter readings are'):
+            ret['title']=''
+            idx-=1
+        else:
+            ret['title']=l.strip()
+        
+        #idx+4 is "Counter readings are offset corrected..."
+        l=lines[idx+5]
+        if not l.startswith('%'):
+            ret['offsets']={}
+            lis=l.split()
+            while len(lis)>0:
+                ret['offsets'][lis.pop()]=float(lis.pop())
+        idx=lines.index('%d')+1
+        ret['columns']=[];
+        while lines[idx].startswith('Col'):
+            ret['columns'].append(lines[idx].split()[2][10:])
+            idx+=1
+        datalines=lines[idx:]
+        for i in range(len(datalines)):
+            datalines[i]=[float(x) for x in datalines[i].split()]
+        ret['data']=np.array(datalines)
+        ret['dataset']=np.array([tuple(a) for a in ret['data'].tolist()], dtype=zip(ret['columns'],[np.double]*len(ret['columns'])))
+        return ret;
+    return None
 def savespheres(spheres,filename):
     """Save sphere structure in a file.
     
@@ -1532,3 +1630,161 @@ def savespheres(spheres,filename):
         filename: filename
     """
     pylab.savetxt(filename,spheres,delimiter='\t')
+def readmask(filename,fieldname=None,dirs='.'):
+    """Try to load a maskfile (matlab(R) matrix file)
+    
+    Inputs:
+        filename: the input file name
+        fieldname: field in the mat file. None to autodetect.
+        dirs: list of directory names to try
+        
+    Outputs:
+        the mask in a numpy array of type np.uint8
+    """
+    if type(dirs)==type(''):
+        dirs=[dirs]
+    f=None
+    for d in dirs:
+        try:
+            f=scipy.io.loadmat(os.path.join(d,filename))
+        except IOError:
+            f=None
+            continue
+        else:
+            break
+    if f is None:
+        raise IOError('Cannot find mask file in any of the given directories!')
+    if fieldname is not None:
+        return f[fieldname].astype(np.uint8)
+    else:
+        validkeys=[k for k in f.keys() if not (k.startswith('_') and k.endswith('_'))];
+        if len(validkeys)<1:
+            raise ValueError('mask file contains no masks!')
+        if len(validkeys)>1:
+            raise ValueError('mask file contains multiple masks!')
+        return f[validkeys[0]].astype(np.uint8)
+def writeparamfile(filename,param):
+    """Write the param structure into a logfile. See writelogfile() for an explanation.
+    
+    Inputs:
+        filename: name of the file.
+        param: param structure (dictionary).
+        
+    Notes:
+        exceptions pass through to the caller.
+    """
+    logfile_dict_float={'FSN':'FSN',
+                        'Sample-to-detector distance (mm)':'Dist',
+                        'Sample thickness (cm)':'Thickness',
+                        'Sample transmission':'Transm',
+                        'Sample position (mm)':'PosSample',
+                        'Temperature':'Temperature',
+                        'Measurement time (sec)':'MeasTime',
+                        'Scattering on 2D detector (photons/sec)':'ScatteringFlux',
+                        'Dark current subtracted (cps)':'dclevel',
+                        'Dark current FSN':'FSNdc',
+                        'Empty beam FSN':'FSNempty',
+                        'Glassy carbon FSN':'FSNref1',
+                        'Glassy carbon thickness (cm)':'Thicknessref1',
+                        'Energy (eV)':'Energy',
+                        'Calibrated energy (eV)':'EnergyCalibrated',
+                        'Calibrated energy':'EnergyCalibrated',
+                        'Beam x y for integration':('BeamPosX','BeamPosY'),
+                        'Normalisation factor (to absolute units)':'NormFactor',
+                        'Relative error of normalisation factor (percentage)':'NormFactorRelativeError',
+                        'Beam size X Y (mm)':('BeamsizeX','BeamsizeY'),
+                        'Pixel size of 2D detector (mm)':'PixelSize',
+                        'Primary intensity at monitor (counts/sec)':'Monitor',
+                        'Primary intensity calculated from GC (photons/sec/mm^2)':'PrimaryIntensity',
+                        'Sample rotation around x axis':'RotXsample',
+                        'Sample rotation around y axis':'RotYsample'
+                        }
+    #this dict. contains the string parameters
+    logfile_dict_str={'Sample name':'Title',
+                      'Sample title':'Title'}
+    #this dict. contains the bool parameters
+    logfile_dict_bool={'Injection between Empty beam and sample measurements?':'InjectionEB',
+                       'Injection between Glassy carbon and sample measurements?':'InjectionGC'
+                       }
+    logfile_dict_list={'FSNs':'FSNs'}
+    logfile_order=['FSN','Title','Dist','Thickness','Transm','PosSample',
+                   'Temperature','MeasTime','ScatteringFlux','dclevel','FSNdc',
+                   'FSNempty','InjectionEB','FSNref1','Thicknessref1',
+                   'InjectionGC','Energy','EnergyCalibrated',('BeamPosX','BeamPosY'),
+                   'NormFactor','NormFactorRelativeError',('BeamsizeX','BeamsizeY'),
+                   'PixelSize','Monitor','PrimaryIntensity','RotXsample','RotYsample']
+    allkeys=param.keys()
+    def boolrepr(x):
+        if x:
+            return 'y'
+        else:
+            return 'n'
+    listrepr=lambda l:' '.join([str(x) for x in l])
+    f=open(filename,'wt')
+    def writekey(k,allkeys=allkeys):
+        #try if it is a float argument
+        k1=[x for x in logfile_dict_float.keys() if logfile_dict_float[x]==k]
+        if len(k1)>0:
+            if type(k)==type(''):
+                k=[k,]
+            k=[x for x in k if x in allkeys]
+            if len(k)==0:
+                return
+            f.write(k1[0])
+            f.write(':\t')
+            f.write(' '.join([str(param[x]) for x in k]))
+            f.write('\n')
+            for x in k:
+                allkeys.remove(x)
+            return
+        #try if it is a str argument
+        k1=[x for x in logfile_dict_str.keys() if logfile_dict_str[x]==k]
+        if len(k1)>0:
+            if type(k)==type(''):
+                k=[k,]
+            k=[x for x in k if x in allkeys]
+            if len(k)==0:
+                return
+            f.write(k1[0])
+            f.write(':\t')
+            f.write(' '.join([str(param[x]) for x in k]))
+            f.write('\n')
+            for x in k:
+                allkeys.remove(x)
+            return
+        #try if it is a bool argument
+        k1=[x for x in logfile_dict_bool.keys() if logfile_dict_bool[x]==k]
+        if len(k1)>0:
+            if type(k)==type(''):
+                k=[k,]
+            k=[x for x in k if x in allkeys]
+            if len(k)==0:
+                return
+            f.write(k1[0])
+            f.write(':\t')
+            f.write(' '.join([boolrepr(param[x]) for x in k]))
+            f.write('\n')
+            for x in k:
+                allkeys.remove(x)
+            return
+        #try if it is a list
+        k1=[x for x in logfile_dict_list.keys() if logfile_dict_list[x]==k]
+        if len(k1)>0:
+            if type(k)==type(''):
+                k=[k,]
+            k=[x for x in k if x in allkeys]
+            if len(k)==0:
+                return
+            f.write(k1[0])
+            f.write(':\t')
+            f.write(' '.join([listrepr(param[x]) for x in k]))
+            f.write('\n')
+            for x in k:
+                allkeys.remove(x)
+            return
+                
+    for k in logfile_order:
+        writekey(k)
+    for k in allkeys:
+        writekey(k)
+    f.close()
\ No newline at end of file
